//     ====================== PerformanceTest.cpp ===========================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 31.07.2008
//     **********************************************************************

#include <stdio.h>
#include "..\..\..\include\AssociativeMemory.h"
#include "highgui.h"

#define cAMFileName "am_example.dat"  // File name of recognition data 
#define cObjMaxNum  4                 // Maximal number of recognition objects

#define cDiagramPointSize 4
#define cBorderSize       70
#define cDiagramYSize     300
#define cPassMaxNum       3

CvAssociativeMemory32S* am;
double PerformanceTime;
int sfrTestPass = 0;

void doSearchObject(CvAM_State aState, long* apData, CvAM_Parameters* apParam, void* args)
{
	static char ObjName[cObjMaxNum][64] = { "Mr. Dmitriy", "Toy \"Heart\"",
		"Book", "Post card"/* "your name of object", ... */};
	char OutNumStr[256];
	
	int i;
	static double bTime,eTime,Tf;
	CvFont font;
	cvInitFont( &font, 2, 1.0, 1.0, 0, 2, 8 );

	IplImage* pImg = (IplImage*)args;

	switch(aState) {
		case cProcessingIsBegun:
			bTime = (double)cvGetTickCount();
			Tf = (double)cvGetTickFrequency()*1000.0;
			break;

		case cRecognizedObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x, apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(255,128,0),1,8,0);
			printf("R   Obj: [%s] ...... Similarity: %7.4f\n", ObjName[*apData-1], apParam->Similarity);
			break;

		case cGeneralizedObject:
		case cTrackedObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x,   apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(255,0,0),2,8,0);

			sprintf(OutNumStr,"%s",ObjName[*apData-1]);
			cvPutText(pImg, OutNumStr, cvPoint(apParam->ObjRect.x+10, apParam->ObjRect.y +
				                               apParam->ObjRect.height+24), &font, CV_RGB(255,255,0));

			printf("G-T Obj: [%s] ...... Similarity: %7.4f\n", ObjName[*apData-1], apParam->Similarity);

			if(apParam->ppUserTrackingInfo) {
				#define cTrcLen 100
				struct TTrcInfo {
					CvPoint TrcPoint[cTrcLen];
					long Age;
				};
				if(!*apParam->ppUserTrackingInfo) {
					*apParam->ppUserTrackingInfo = new TTrcInfo;
					memset(*apParam->ppUserTrackingInfo, 0, sizeof(TTrcInfo));
				} else {
					TTrcInfo* pTrcInfo = (TTrcInfo*)*apParam->ppUserTrackingInfo;
					memcpy(&pTrcInfo->TrcPoint[0], &pTrcInfo->TrcPoint[1], sizeof(CvPoint)*(cTrcLen-1));
					pTrcInfo->TrcPoint[cTrcLen-1] = cvPoint(apParam->ObjRect.x + (apParam->ObjRect.width>>1),
						                                    apParam->ObjRect.y + (apParam->ObjRect.height>>1));
					if(pTrcInfo->Age < cTrcLen) { pTrcInfo->Age++; }

					if(pTrcInfo->Age > 2) {
						for(i=1; i<pTrcInfo->Age; i++) {
							cvLine(pImg, pTrcInfo->TrcPoint[cTrcLen-i-1],
								         pTrcInfo->TrcPoint[cTrcLen-i], cvScalar(255,255,0),2,8,0);
						}
						cvCircle(pImg, pTrcInfo->TrcPoint[cTrcLen-1],4,cvScalar(0,255,0),2,8,0);
					}
				}
			}
			break;

		case cLearnThisObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x,   apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x +  apParam->ObjRect.width,
								       apParam->ObjRect.y +  apParam->ObjRect.height),
									   cvScalar(255,255,0),2,8,0);
			printf("L   Obj: [%s] ...... Similarity: %7.4f, Bases: %lu\n",
				   ObjName[*apData-1], apParam->Similarity, apParam->TotalABases);
			break;

		case cProcessingIsFinished:
			eTime = (double)cvGetTickCount(); PerformanceTime = (eTime-bTime)/Tf;
			if(!sfrTestPass) { sprintf(OutNumStr,"%5.1f ms",PerformanceTime); } else {
				sprintf(OutNumStr,"%5.1f ms Fail:%5.2f%%", PerformanceTime, am->GetSearchFailRate()*100.0);
				cvPutText(pImg, (sfrTestPass == 1)?"Advanced searching":"Quick searching",
						  cvPoint(0, 50), &font, CV_RGB(0,255,0));
			}
			cvPutText(pImg, OutNumStr, cvPoint(0, 24), &font, CV_RGB(0,255,0));
			break;
	}
}

char* TimeToStr(long aPOS_MSEC)
{
	static char TimeStr[64];
	int hour    = int(aPOS_MSEC/(3600000));
	aPOS_MSEC  %= 3600000L;
	int minute  = int(aPOS_MSEC/(60000));
	aPOS_MSEC  %= 60000L;
	int second  = int(aPOS_MSEC/1000);
	int millisecond = int(aPOS_MSEC % 1000);
	sprintf(TimeStr, "%2.2d:%2.2d:%2.2d.%3.3d", hour, minute, second, millisecond);
	return TimeStr;
}

int main(int argc, char* argv[])
{
	int       i;
	ULONG     FilmLengthMs  = 0;
	double    FilmFrameRate;
	IplImage* OutImg;
	static CvScalar ColorForPass[cPassMaxNum] = { {0,0,255}, {0,255,255}, {255,255,0} };

	am = new CvAssociativeMemory32S;
	
	if(argc <= 1 || (argc == 2 && strstr(argv[1],"/?"))) {
		printf("\n\n=== Performance test help ===\n\n"
			   "PerformanceTest.exe [file.avi]\n"
               "Example:\n\n"
			   "PerformanceTest.exe RcgTraining4s.avi\n\n");
		return 0;
	}

	char* argv1 = _strlwr(argv[1]);
	if(!strstr(argv1,".avi")) {
		printf("AVI files are processing only!\n\n");
		return -1;
	}

	#define cStrMax 256
	char FileName[cStrMax];
	char OutStr[cStrMax];
	char CommandsStr[cStrMax];
	memset(CommandsStr,0,cStrMax);
	for(i=2; i<argc; i++) strcat(CommandsStr,argv[i]);
	memset(FileName,0,cStrMax);
	CvCapture*     InAVI;
    IplImage*      frame = 0;
	IplImage*      gray  = 0;

	int WIDTH, HEIGHT;
	CvSize ImgSize;

	printf("Processing avi file...\n");
	memcpy(FileName,argv1,strstr(argv1,".avi")-argv1);
	InAVI = cvCaptureFromFile(argv1);
	if(!InAVI) {
		printf("Can't find file %s.avi\n",FileName);
		return -1;
	}
	cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,1);
	FilmLengthMs  = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
	cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,0);
	WIDTH         = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_WIDTH));
	HEIGHT        = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_HEIGHT));
	FilmFrameRate = cvGetCaptureProperty(InAVI,CV_CAP_PROP_FPS);
	ImgSize       = cvSize(WIDTH, HEIGHT);

	OutImg = cvCreateImage( ImgSize, 8, 3 );
	gray = cvCreateImage( ImgSize, 8, 1 );

	if(am->Load(cAMFileName)) {
		printf("\n  Load associative bases from file: %s\n"
		   "Associative bases ......... %-7lu\n"
		   "Associative Levels ........ %-7u\n"
		   "Current base index ........ %-7I64u\n"
		   "Wr/Rd counter ............. %-7I64u\n\n",
		   cAMFileName, am->GetTotalABases(),
		   am->GetTotalLevels(), am->GetCurIndex(), am->GetWrRdCounter());
	} else {
			printf("Can't open file %s\n",cAMFileName);
			return -1;
	}
	int TotalPoints = 100;
	int PointInterval = int((FilmLengthMs*FilmFrameRate)/(1000*TotalPoints));
	int ResImgW = TotalPoints*cDiagramPointSize + cBorderSize*2;
	int ResImgH = cDiagramYSize + cBorderSize*2;
	CvFont BigFont;
	cvInitFont( &BigFont, 2, 0.8, 0.8, 0, 2, 8 );
	CvFont MiddleFont;
	cvInitFont( &MiddleFont, 2, 0.7, 0.7, 0.3, 1, 8 );
	CvFont SmallFont;
	cvInitFont( &SmallFont, 1, 1.0, 1.0, 0, 1, 8 );

	IplImage* ResImage = cvCreateImage(cvSize(ResImgW, ResImgH), IPL_DEPTH_8U, 3);
	cvSet(ResImage,cvScalarAll(180),0);

	int DiagramXBegin    = cBorderSize;
	int DiagramXEnd      = ResImgW-cBorderSize;
	int DiagramYTop      = ResImgH-(cDiagramYSize + cBorderSize);
	int DiagramYMiddle   = ResImgH-(cDiagramYSize/2 + cBorderSize);
	int DiagramYBottom   = ResImgH-cBorderSize;
	int DiagramYBottomT1 = DiagramYBottom + 16;
	int DiagramYBottomT2 = DiagramYBottom + 40;
	
	cvLine(ResImage,cvPoint(DiagramXBegin, DiagramYBottom),
		            cvPoint(DiagramXEnd,   DiagramYBottom), CV_RGB(0,0,0),1,8,0);
	cvLine(ResImage,cvPoint(DiagramXBegin, DiagramYMiddle),
		            cvPoint(DiagramXEnd,   DiagramYMiddle), CV_RGB(0,0,0),1,8,0);
	cvLine(ResImage,cvPoint(DiagramXBegin, DiagramYTop),
		            cvPoint(DiagramXEnd,   DiagramYTop),    CV_RGB(0,0,0),1,8,0);
	cvLine(ResImage,cvPoint(DiagramXBegin, DiagramYTop),
		            cvPoint(DiagramXBegin, DiagramYBottom), CV_RGB(0,0,0),1,8,0);
	cvLine(ResImage,cvPoint(DiagramXEnd,   DiagramYTop),
		            cvPoint(DiagramXEnd,   DiagramYBottom), CV_RGB(0,0,0),1,8,0);

	cvPutText(ResImage, "Diagram of search performance", cvPoint(60, DiagramYTop-20), &BigFont, CV_RGB(0,0,0));
	cvPutText(ResImage, "0.0", cvPoint(2, DiagramYBottom), &MiddleFont, CV_RGB(0,0,0));
	cvPutText(ResImage, "30ms", cvPoint(2, DiagramYMiddle), &MiddleFont, CV_RGB(0,0,0));
	cvPutText(ResImage, "60ms", cvPoint(2, DiagramYTop),    &MiddleFont, CV_RGB(0,0,0));
	sprintf(OutStr, "%35s.avi   %s", FileName, TimeToStr(FilmLengthMs));
	cvPutText(ResImage, OutStr, cvPoint(DiagramXBegin, DiagramYBottomT1),&SmallFont, CV_RGB(0,0,0));
	cvPutText(ResImage, "Advanced,", cvPoint(DiagramXBegin, DiagramYBottomT2),&SmallFont, ColorForPass[0]);
	cvPutText(ResImage, "quick,", cvPoint(DiagramXBegin+90, DiagramYBottomT2),&SmallFont, ColorForPass[1]);
	cvPutText(ResImage, "linear", cvPoint(DiagramXBegin+150, DiagramYBottomT2),&SmallFont, ColorForPass[2]);
	cvPutText(ResImage, "searching", cvPoint(DiagramXBegin+204, DiagramYBottomT2),&SmallFont, CV_RGB(0,0,0));
	

	cvNamedWindow("Diagram", 1);
	cvNamedWindow("Result", 0);

	am->SetSearchMode(cAdvancedSearching);


	int IntervalCounter = 0;
	int PointNum = 0;
	CvPoint PrevPoint, NextPoint;

	for(int Pass=0;;) {
		
		if(!cvGrabFrame(InAVI)) {
			am->Load(cAMFileName);
			cvSetCaptureProperty( InAVI, CV_CAP_PROP_POS_MSEC, 0 );
			PointNum = 0;
			am->SetSearchMode(Pass?cLinearSearching:cQuickSearching);
			if(++Pass >= cPassMaxNum) {
				am->SetSearchMode(sfrTestPass?cQuickSearchingTestMode:cAdvancedSearchingTestMode);
				if(++sfrTestPass > 2) { break; } else { continue; }
			} else { continue; }
		} 
	
		frame = cvRetrieveFrame(InAVI);

		cvFlip(frame);

		cvCopy(frame, OutImg);
		
		cvCvtColor(frame, gray, CV_BGR2GRAY);


		am->SetImage(gray);
		am->ObjectTracking(doSearchObject, OutImg);

		printf("==> film current position: %s\n",
			   TimeToStr(long(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC))));

	    cvShowImage("Result", OutImg);

		IntervalCounter = (++IntervalCounter) % PointInterval;

		if(!sfrTestPass && !IntervalCounter) {
			NextPoint = cvPoint(PointNum*cDiagramPointSize + cBorderSize,
								ResImgH - cBorderSize - int(cDiagramYSize*(PerformanceTime/60.0)));
			if(PointNum) {
				cvLine(ResImage, PrevPoint, NextPoint, ColorForPass[Pass], cDiagramPointSize,8,0);
			}
			PrevPoint = NextPoint;
			PointNum++;
		}
		cvShowImage("Diagram", ResImage);
		
		if( cvWaitKey(10) >= 0 ) break;
	}

	delete am;

	sprintf(FileName,"Performance_Test_%u_ms_%ux%u.jpg", FilmLengthMs, WIDTH, HEIGHT);
	if(cvSaveImage(FileName, ResImage)) {
		printf("\nResult image was saved into file %s\n\n", FileName);
	}
	
	cvReleaseImage(&ResImage);
	cvReleaseCapture(&InAVI);
	cvReleaseImage(&gray);
	
	return 0;
}